

#include "tekterminal.h"

namespace fabgl {

// Terminal identification ID
// 64 = VT420
// 1 = support for 132 columns
// 6 = selective erase
// 22 = color
const char TERMID[] = "?64;1;6;22c";
#define ISCTRLCHAR(c) ((c) <= ASCII_US || (c) == ASCII_DEL)
// Map "DEC Special Graphics Character Set" to CP437
static const uint8_t DECGRAPH_TO_CP437[255] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,
                                               26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
                                               50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73,
                                               74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94,
                                               /* 95 */ 32, /* 96 */ 4, /* 97 */ 177, /* 98 (not implemented) */ 63, /* 99 (not implemented) */ 63,
                                               /* 100 (not implemented) */ 63, /* 101 (not implemented) */ 63, /* 102 */ 248, /* 103 */ 241,
                                               /* 104 (not implemented) */ 63, /* 105 (not implemented) */ 63, /* 106 */ 217, /* 107 */ 191, /* 108 */ 218,
                                               /* 109 */ 192, /* 110 */ 197, /* 111 (not implemented) */ 63, /* 112 (not implemented) */ 63, /* 113 */ 196,
                                               /* 114 (not implemented) */ 63, /* 115 (not implemented) */ 63, /* 116 */ 195, /* 117 */ 180, /* 118 */ 193,
                                               /* 119 */ 194, /* 120 */ 179, /* 121 */ 243, /* 122 */ 242, /* 123 */ 227, /* 124 (not implemented) */ 63,
                                               /* 125 */ 156, /* 126 */ 249};

/*
 * 
 * Hershey SIMPLEX font
 * origin: at upper left of cell. 
 * x positive to the right
 * y positive down
 *
 * line
 *   0:  +---------+   [[[[    ---
 *       |             [       0 ... 3 = 4
 *   4:  +     A       [       --- 
 *       |    A A      [
 *       |   A   A     [       4 ... 25 = 22 nominal character height
 *       |  AAAAAAA    [        
 *  25:  + A       A   [       --- base line
 *       |             [       26...32 = 8
 *  32:  +---------+   [[[[    --- 
 * 
 * character height = 22
 * cell height: 33 (then '()[]{}/|[' on subsequent lines just touch)
 * cell with:  [8 ... 30]
 * sizeof(font data) = 760 bytes
 * *
 * * Compact storage version Martin Hepperle, 2018
 */

typedef unsigned char BYTE;

typedef struct
{
   BYTE width;
   BYTE points;
   const char * code;
} TEKCHAR;


TEKCHAR m_charset[]=
{
   { /* ASCII 32 ' '*/
      16, 0,
     ""
   },
   { /* ASCII 33 '!'*/
      10, 15,
     "\x05\x04\x05\x12\x7f\x05\x17\x04\x18\x05\x19\x06\x18"
      "\x05\x17"
   },
   { /* ASCII 34 '"'*/
      16, 9,
     "\x04\x04\x04\x0b\x7f\x0c\x04\x0c\x0b"
   },
   { /* ASCII 35 '#'*/
      21, 19,
     "\x0b\x00\x04\x20\x7f\x11\x00\x0a\x20\x7f\x04\x0d"
      "\x12\x0d\x7f\x03\x13\x11\x13"
   },
   { /* ASCII 36 '$'*/
      20, 50,
     "\x08\x00\x08\x1d\x7f\x0c\x00\x0c\x1d\x7f\x11\x07"
      "\x0f\x05\x0c\x04\x08\x04\x05\x05\x03\x07\x03\x09\x04\x0b\x05\x0c"
      "\x07\x0d\x0d\x0f\x0f\x10\x10\x11\x11\x13\x11\x16\x0f\x18\x0c\x19"
      "\x08\x19\x05\x18\x03\x16"
   },
   { /* ASCII 37 '%'*/
      24, 60,
     "\x15\x04\x03\x19\x7f\x08\x04\x0a\x06\x0a\x08\x09\x0a"
      "\x07\x0b\x05\x0b\x03\x09\x03\x07\x04\x05\x06\x04\x08\x04\x0a\x05"
      "\x0d\x06\x10\x06\x13\x05\x15\x04\x7f\x11\x12\x0f\x13\x0e\x15"
      "\x0e\x17\x10\x19\x12\x19\x14\x18\x15\x16\x15\x14\x13\x12\x11\x12"
   },
   { /* ASCII 38 '&'*/
      26, 68,
     "\x17\x0d\x17\x0c\x16\x0b\x15\x0b\x14\x0c\x13\x0e\x11\x13"
      "\x0f\x16\x0d\x18\x0b\x19\x07\x19\x05\x18\x04\x17\x03\x15\x03\x13"
      "\x04\x11\x05\x10\x0c\x0c\x0d\x0b\x0e\x09\x0e\x07\x0d\x05\x0b\x04"
      "\x09\x05\x08\x07\x08\x09\x09\x0c\x0b\x0f\x10\x16\x12\x18\x14\x19"
      "\x16\x19\x17\x18\x17\x17"
   },
   { /* ASCII 39 '''*/
      10, 14,
     "\x05\x06\x04\x05\x05\x04\x06\x05\x06\x07\x05\x09\x04\x0a"
   },
   { /* ASCII 40 '('*/
      14, 20,
     "\x0b\x00\x09\x02\x07\x05\x05\x09\x04\x0e\x04\x12\x05\x17"
      "\x07\x1b\x09\x1e\x0b\x20"
   },
   { /* ASCII 41 ')'*/
      14, 20,
     "\x03\x00\x05\x02\x07\x05\x09\x09\x0a\x0e\x0a\x12\x09\x17"
      "\x07\x1b\x05\x1e\x03\x20"
   },
   { /* ASCII 42 '*'*/
      16, 14,
     "\x08\x04\x08\x10\x7f\x03\x07\x0d\x0d\x7f\x0d\x07"
      "\x03\x0d"
   },
   { /* ASCII 43 '+'*/
      26, 9,
     "\x0d\x07\x0d\x19\x7f\x04\x10\x16\x10"
   },
   { /* ASCII 44 ','*/
      10, 16,
     "\x06\x18\x05\x19\x04\x18\x05\x17\x06\x18\x06\x1a\x05\x1c"
      "\x04\x1d"
   },
   { /* ASCII 45 '-'*/
      26, 4,
     "\x04\x10\x16\x10"
   },
   { /* ASCII 46 '.'*/
      10, 10,
     "\x05\x17\x04\x18\x05\x19\x06\x18\x05\x17"
   },
   { /* ASCII 47 '/'*/
      22, 4,
     "\x14\x00\x02\x20"
   },
   { /* ASCII 48 '0'*/
      20, 34,
     "\x09\x04\x06\x05\x04\x08\x03\x0d\x03\x10\x04\x15\x06\x18"
      "\x09\x19\x0b\x19\x0e\x18\x10\x15\x11\x10\x11\x0d\x10\x08\x0e\x05"
      "\x0b\x04\x09\x04"
   },
   { /* ASCII 49 '1'*/
      20, 8,
     "\x06\x08\x08\x07\x0b\x04\x0b\x19"
   },
   { /* ASCII 50 '2'*/
      20, 28,
     "\x04\x09\x04\x08\x05\x06\x06\x05\x08\x04\x0c\x04\x0e\x05"
      "\x0f\x06\x10\x08\x10\x0a\x0f\x0c\x0d\x0f\x03\x19\x11\x19"
   },
   { /* ASCII 51 '3'*/
      20, 30,
     "\x05\x04\x10\x04\x0a\x0c\x0d\x0c\x0f\x0d\x10\x0e\x11\x11"
      "\x11\x13\x10\x16\x0e\x18\x0b\x19\x08\x19\x05\x18\x04\x17\x03\x15"
   },
   { /* ASCII 52 '4'*/
      20, 11,
     "\x0d\x04\x03\x12\x12\x12\x7f\x0d\x04\x0d\x19"
   },
   { /* ASCII 53 '5'*/
      20, 34,
     "\x0f\x04\x05\x04\x04\x0d\x05\x0c\x08\x0b\x0b\x0b\x0e\x0c"
      "\x10\x0e\x11\x11\x11\x13\x10\x16\x0e\x18\x0b\x19\x08\x19\x05\x18"
      "\x04\x17\x03\x15"
   },
   { /* ASCII 54 '6'*/
      20, 46,
     "\x10\x07\x0f\x05\x0c\x04\x0a\x04\x07\x05\x05\x08\x04\x0d"
      "\x04\x12\x05\x16\x07\x18\x0a\x19\x0b\x19\x0e\x18\x10\x16\x11\x13"
      "\x11\x12\x10\x0f\x0e\x0d\x0b\x0c\x0a\x0c\x07\x0d\x05\x0f\x04\x12"
   },
   { /* ASCII 55 '7'*/
      20, 9,
     "\x11\x04\x07\x19\x7f\x03\x04\x11\x04"
   },
   { /* ASCII 56 '8'*/
      20, 58,
     "\x08\x04\x05\x05\x04\x07\x04\x09\x05\x0b\x07\x0c\x0b\x0d"
      "\x0e\x0e\x10\x10\x11\x12\x11\x15\x10\x17\x0f\x18\x0c\x19\x08\x19"
      "\x05\x18\x04\x17\x03\x15\x03\x12\x04\x10\x06\x0e\x09\x0d\x0d\x0c"
      "\x0f\x0b\x10\x09\x10\x07\x0f\x05\x0c\x04\x08\x04"
   },
   { /* ASCII 57 '9'*/
      20, 46,
     "\x10\x0b\x0f\x0e\x0d\x10\x0a\x11\x09\x11\x06\x10\x04\x0e"
      "\x03\x0b\x03\x0a\x04\x07\x06\x05\x09\x04\x0a\x04\x0d\x05\x0f\x07"
      "\x10\x0b\x10\x10\x0f\x15\x0d\x18\x0a\x19\x08\x19\x05\x18\x04\x16"
   },
   { /* ASCII 58 ':'*/
      10, 21,
     "\x05\x0b\x04\x0c\x05\x0d\x06\x0c\x05\x0b\x7f\x05\x17"
      "\x04\x18\x05\x19\x06\x18\x05\x17"
   },
   { /* ASCII 59 ';'*/
      10, 27,
     "\x05\x0b\x04\x0c\x05\x0d\x06\x0c\x05\x0b\x7f\x06\x18"
      "\x05\x19\x04\x18\x05\x17\x06\x18\x06\x1a\x05\x1c\x04\x1d"
   },
   { /* ASCII 60 '<'*/
      24, 6,
     "\x14\x07\x04\x10\x14\x19"
   },
   { /* ASCII 61 '='*/
      26, 9,
     "\x04\x0d\x16\x0d\x7f\x04\x13\x16\x13"
   },
   { /* ASCII 62 '>'*/
      24, 6,
     "\x04\x07\x14\x10\x04\x19"
   },
   { /* ASCII 63 '?'*/
      18, 39,
     "\x03\x09\x03\x08\x04\x06\x05\x05\x07\x04\x0b\x04\x0d\x05"
      "\x0e\x06\x0f\x08\x0f\x0a\x0e\x0c\x0d\x0d\x09\x0f\x09\x12\x7f"
      "\x09\x17\x08\x18\x09\x19\x0a\x18\x09\x17"
   },
   { /* ASCII 64 '@'*/
      27, 107,
     "\x12\x0c\x11\x0a\x0f\x09\x0c\x09\x0a\x0a\x09\x0b\x08\x0e"
      "\x08\x11\x09\x13\x0b\x14\x0e\x14\x10\x13\x11\x11\x7f\x0c\x09"
      "\x0a\x0b\x09\x0e\x09\x11\x0a\x13\x0b\x14\x7f\x12\x09\x11\x11"
      "\x11\x13\x13\x14\x15\x14\x17\x12\x18\x0f\x18\x0d\x17\x0a\x16\x08"
      "\x14\x06\x12\x05\x0f\x04\x0c\x04\x09\x05\x07\x06\x05\x08\x04\x0a"
      "\x03\x0d\x03\x10\x04\x13\x05\x15\x07\x17\x09\x18\x0c\x19\x0f\x19"
      "\x12\x18\x14\x17\x15\x16\x7f\x13\x09\x12\x11\x12\x13\x13\x14"
   },
   { /* ASCII 65 'A'*/
      18, 14,
     "\x09\x04\x01\x19\x7f\x09\x04\x11\x19\x7f\x04\x12"
      "\x0e\x12"
   },
   { /* ASCII 66 'B'*/
      21, 44,
     "\x04\x04\x04\x19\x7f\x04\x04\x0d\x04\x10\x05\x11\x06"
      "\x12\x08\x12\x0a\x11\x0c\x10\x0d\x0d\x0e\x7f\x04\x0e\x0d\x0e"
      "\x10\x0f\x11\x10\x12\x12\x12\x15\x11\x17\x10\x18\x0d\x19\x04\x19"
   },
   { /* ASCII 67 'C'*/
      21, 36,
     "\x12\x09\x11\x07\x0f\x05\x0d\x04\x09\x04\x07\x05\x05\x07"
      "\x04\x09\x03\x0c\x03\x11\x04\x14\x05\x16\x07\x18\x09\x19\x0d\x19"
      "\x0f\x18\x11\x16\x12\x14"
   },
   { /* ASCII 68 'D'*/
      21, 29,
     "\x04\x04\x04\x19\x7f\x04\x04\x0b\x04\x0e\x05\x10\x07"
      "\x11\x09\x12\x0c\x12\x11\x11\x14\x10\x16\x0e\x18\x0b\x19\x04\x19"
   },
   { /* ASCII 69 'E'*/
      19, 19,
     "\x04\x04\x04\x19\x7f\x04\x04\x11\x04\x7f\x04\x0e"
      "\x0c\x0e\x7f\x04\x19\x11\x19"
   },
   { /* ASCII 70 'F'*/
      18, 14,
     "\x04\x04\x04\x19\x7f\x04\x04\x11\x04\x7f\x04\x0e"
      "\x0c\x0e"
   },
   { /* ASCII 71 'G'*/
      21, 43,
     "\x12\x09\x11\x07\x0f\x05\x0d\x04\x09\x04\x07\x05\x05\x07"
      "\x04\x09\x03\x0c\x03\x11\x04\x14\x05\x16\x07\x18\x09\x19\x0d\x19"
      "\x0f\x18\x11\x16\x12\x14\x12\x11\x7f\x0d\x11\x12\x11"
   },
   { /* ASCII 72 'H'*/
      22, 14,
     "\x04\x04\x04\x19\x7f\x12\x04\x12\x19\x7f\x04\x0e"
      "\x12\x0e"
   },
   { /* ASCII 73 'I'*/
      8, 4,
     "\x04\x04\x04\x19"
   },
   { /* ASCII 74 'J'*/
      16, 20,
     "\x0c\x04\x0c\x14\x0b\x17\x0a\x18\x08\x19\x06\x19\x04\x18"
      "\x03\x17\x02\x14\x02\x12"
   },
   { /* ASCII 75 'K'*/
      21, 14,
     "\x04\x04\x04\x19\x7f\x12\x04\x04\x12\x7f\x09\x0d"
      "\x12\x19"
   },
   { /* ASCII 76 'L'*/
      17, 9,
     "\x04\x04\x04\x19\x7f\x04\x19\x10\x19"
   },
   { /* ASCII 77 'M'*/
      24, 19,
     "\x04\x04\x04\x19\x7f\x04\x04\x0c\x19\x7f\x14\x04"
      "\x0c\x19\x7f\x14\x04\x14\x19"
   },
   { /* ASCII 78 'N'*/
      22, 14,
     "\x04\x04\x04\x19\x7f\x04\x04\x12\x19\x7f\x12\x04"
      "\x12\x19"
   },
   { /* ASCII 79 'O'*/
      22, 42,
     "\x09\x04\x07\x05\x05\x07\x04\x09\x03\x0c\x03\x11\x04\x14"
      "\x05\x16\x07\x18\x09\x19\x0d\x19\x0f\x18\x11\x16\x12\x14\x13\x11"
      "\x13\x0c\x12\x09\x11\x07\x0f\x05\x0d\x04\x09\x04"
   },
   { /* ASCII 80 'P'*/
      21, 25,
     "\x04\x04\x04\x19\x7f\x04\x04\x0d\x04\x10\x05\x11\x06"
      "\x12\x08\x12\x0b\x11\x0d\x10\x0e\x0d\x0f\x04\x0f"
   },
   { /* ASCII 81 'Q'*/
      22, 47,
     "\x09\x04\x07\x05\x05\x07\x04\x09\x03\x0c\x03\x11\x04\x14"
      "\x05\x16\x07\x18\x09\x19\x0d\x19\x0f\x18\x11\x16\x12\x14\x13\x11"
      "\x13\x0c\x12\x09\x11\x07\x0f\x05\x0d\x04\x09\x04\x7f\x0c\x15"
      "\x12\x1b"
   },
   { /* ASCII 82 'R'*/
      21, 30,
     "\x04\x04\x04\x19\x7f\x04\x04\x0d\x04\x10\x05\x11\x06"
      "\x12\x08\x12\x0a\x11\x0c\x10\x0d\x0d\x0e\x04\x0e\x7f\x0b\x0e"
      "\x12\x19"
   },
   { /* ASCII 83 'S'*/
      20, 40,
     "\x11\x07\x0f\x05\x0c\x04\x08\x04\x05\x05\x03\x07\x03\x09"
      "\x04\x0b\x05\x0c\x07\x0d\x0d\x0f\x0f\x10\x10\x11\x11\x13\x11\x16"
      "\x0f\x18\x0c\x19\x08\x19\x05\x18\x03\x16"
   },
   { /* ASCII 84 'T'*/
      16, 9,
     "\x08\x04\x08\x19\x7f\x01\x04\x0f\x04"
   },
   { /* ASCII 85 'U'*/
      22, 20,
     "\x04\x04\x04\x13\x05\x16\x07\x18\x0a\x19\x0c\x19\x0f\x18"
      "\x11\x16\x12\x13\x12\x04"
   },
   { /* ASCII 86 'V'*/
      18, 9,
     "\x01\x04\x09\x19\x7f\x11\x04\x09\x19"
   },
   { /* ASCII 87 'W'*/
      24, 19,
     "\x02\x04\x07\x19\x7f\x0c\x04\x07\x19\x7f\x0c\x04"
      "\x11\x19\x7f\x16\x04\x11\x19"
   },
   { /* ASCII 88 'X'*/
      20, 9,
     "\x03\x04\x11\x19\x7f\x11\x04\x03\x19"
   },
   { /* ASCII 89 'Y'*/
      18, 11,
     "\x01\x04\x09\x0e\x09\x19\x7f\x11\x04\x09\x0e"
   },
   { /* ASCII 90 'Z'*/
      20, 14,
     "\x11\x04\x03\x19\x7f\x03\x04\x11\x04\x7f\x03\x19"
      "\x11\x19"
   },
   { /* ASCII 91 '['*/
      14, 19,
     "\x04\x00\x04\x20\x7f\x05\x00\x05\x20\x7f\x04\x00"
      "\x0b\x00\x7f\x04\x20\x0b\x20"
   },
   { /* ASCII 92 '\'*/
      14, 4,
     "\x00\x04\x0e\x1c"
   },
   { /* ASCII 93 ']'*/
      14, 19,
     "\x09\x00\x09\x20\x7f\x0a\x00\x0a\x20\x7f\x03\x00"
      "\x0a\x00\x7f\x03\x20\x0a\x20"
   },
   { /* ASCII 94 '^'*/
      16, 18,
     "\x06\x0a\x08\x07\x0a\x0a\x7f\x03\x0d\x08\x08\x0d\x0d"
      "\x7f\x08\x08\x08\x19"
   },
   { /* ASCII 95 '_'*/
      16, 4,
     "\x00\x1b\x10\x1b"
   },
   { /* ASCII 96 '`'*/
      10, 14,
     "\x06\x04\x05\x05\x04\x07\x04\x09\x05\x0a\x06\x09\x05\x08"
   },
   { /* ASCII 97 'a'*/
      19, 33,
     "\x0f\x0b\x0f\x19\x7f\x0f\x0e\x0d\x0c\x0b\x0b\x08\x0b"
      "\x06\x0c\x04\x0e\x03\x11\x03\x13\x04\x16\x06\x18\x08\x19\x0b\x19"
      "\x0d\x18\x0f\x16"
   },
   { /* ASCII 98 'b'*/
      19, 33,
     "\x04\x04\x04\x19\x7f\x04\x0e\x06\x0c\x08\x0b\x0b\x0b"
      "\x0d\x0c\x0f\x0e\x10\x11\x10\x13\x0f\x16\x0d\x18\x0b\x19\x08\x19"
      "\x06\x18\x04\x16"
   },
   { /* ASCII 99 'c'*/
      18, 28,
     "\x0f\x0e\x0d\x0c\x0b\x0b\x08\x0b\x06\x0c\x04\x0e\x03\x11"
      "\x03\x13\x04\x16\x06\x18\x08\x19\x0b\x19\x0d\x18\x0f\x16"
   },
   { /* ASCII 100 'd'*/
      19, 33,
     "\x0f\x04\x0f\x19\x7f\x0f\x0e\x0d\x0c\x0b\x0b\x08\x0b"
      "\x06\x0c\x04\x0e\x03\x11\x03\x13\x04\x16\x06\x18\x08\x19\x0b\x19"
      "\x0d\x18\x0f\x16"
   },
   { /* ASCII 101 'e'*/
      18, 34,
     "\x03\x11\x0f\x11\x0f\x0f\x0e\x0d\x0d\x0c\x0b\x0b\x08\x0b"
      "\x06\x0c\x04\x0e\x03\x11\x03\x13\x04\x16\x06\x18\x08\x19\x0b\x19"
      "\x0d\x18\x0f\x16"
   },
   { /* ASCII 102 'f'*/
      12, 15,
     "\x0a\x04\x08\x04\x06\x05\x05\x08\x05\x19\x7f\x02\x0b"
      "\x09\x0b"
   },
   { /* ASCII 103 'g'*/
      19, 43,
     "\x0f\x0b\x0f\x1b\x0e\x1e\x0d\x1f\x0b\x20\x08\x20\x06\x1f"
      "\x7f\x0f\x0e\x0d\x0c\x0b\x0b\x08\x0b\x06\x0c\x04\x0e\x03\x11"
      "\x03\x13\x04\x16\x06\x18\x08\x19\x0b\x19\x0d\x18\x0f\x16"
   },
   { /* ASCII 104 'h'*/
      19, 19,
     "\x04\x04\x04\x19\x7f\x04\x0f\x07\x0c\x09\x0b\x0c\x0b"
      "\x0e\x0c\x0f\x0f\x0f\x19"
   },
   { /* ASCII 105 'i'*/
      8, 15,
     "\x03\x04\x04\x05\x05\x04\x04\x03\x03\x04\x7f\x04\x0b"
      "\x04\x19"
   },
   { /* ASCII 106 'j'*/
      10, 21,
     "\x05\x04\x06\x05\x07\x04\x06\x03\x05\x04\x7f\x06\x0b"
      "\x06\x1c\x05\x1f\x03\x20\x01\x20"
   },
   { /* ASCII 107 'k'*/
      17, 14,
     "\x04\x04\x04\x19\x7f\x0e\x0b\x04\x15\x7f\x08\x11"
      "\x0f\x19"
   },
   { /* ASCII 108 'l'*/
      8, 4,
     "\x04\x04\x04\x19"
   },
   { /* ASCII 109 'm'*/
      30, 34,
     "\x04\x0b\x04\x19\x7f\x04\x0f\x07\x0c\x09\x0b\x0c\x0b"
      "\x0e\x0c\x0f\x0f\x0f\x19\x7f\x0f\x0f\x12\x0c\x14\x0b\x17\x0b"
      "\x19\x0c\x1a\x0f\x1a\x19"
   },
   { /* ASCII 110 'n'*/
      19, 19,
     "\x04\x0b\x04\x19\x7f\x04\x0f\x07\x0c\x09\x0b\x0c\x0b"
      "\x0e\x0c\x0f\x0f\x0f\x19"
   },
   { /* ASCII 111 'o'*/
      19, 34,
     "\x08\x0b\x06\x0c\x04\x0e\x03\x11\x03\x13\x04\x16\x06\x18"
      "\x08\x19\x0b\x19\x0d\x18\x0f\x16\x10\x13\x10\x11\x0f\x0e\x0d\x0c"
      "\x0b\x0b\x08\x0b"
   },
   { /* ASCII 112 'p'*/
      19, 33,
     "\x04\x0b\x04\x20\x7f\x04\x0e\x06\x0c\x08\x0b\x0b\x0b"
      "\x0d\x0c\x0f\x0e\x10\x11\x10\x13\x0f\x16\x0d\x18\x0b\x19\x08\x19"
      "\x06\x18\x04\x16"
   },
   { /* ASCII 113 'q'*/
      19, 33,
     "\x0f\x0b\x0f\x20\x7f\x0f\x0e\x0d\x0c\x0b\x0b\x08\x0b"
      "\x06\x0c\x04\x0e\x03\x11\x03\x13\x04\x16\x06\x18\x08\x19\x0b\x19"
      "\x0d\x18\x0f\x16"
   },
   { /* ASCII 114 'r'*/
      13, 15,
     "\x04\x0b\x04\x19\x7f\x04\x11\x05\x0e\x07\x0c\x09\x0b"
      "\x0c\x0b"
   },
   { /* ASCII 115 's'*/
      17, 34,
     "\x0e\x0e\x0d\x0c\x0a\x0b\x07\x0b\x04\x0c\x03\x0e\x04\x10"
      "\x06\x11\x0b\x12\x0d\x13\x0e\x15\x0e\x16\x0d\x18\x0a\x19\x07\x19"
      "\x04\x18\x03\x16"
   },
   { /* ASCII 116 't'*/
      12, 15,
     "\x05\x04\x05\x15\x06\x18\x08\x19\x0a\x19\x7f\x02\x0b"
      "\x09\x0b"
   },
   { /* ASCII 117 'u'*/
      19, 19,
     "\x04\x0b\x04\x15\x05\x18\x07\x19\x0a\x19\x0c\x18\x0f\x15"
      "\x7f\x0f\x0b\x0f\x19"
   },
   { /* ASCII 118 'v'*/
      16, 9,
     "\x02\x0b\x08\x19\x7f\x0e\x0b\x08\x19"
   },
   { /* ASCII 119 'w'*/
      22, 19,
     "\x03\x0b\x07\x19\x7f\x0b\x0b\x07\x19\x7f\x0b\x0b"
      "\x0f\x19\x7f\x13\x0b\x0f\x19"
   },
   { /* ASCII 120 'x'*/
      17, 9,
     "\x03\x0b\x0e\x19\x7f\x0e\x0b\x03\x19"
   },
   { /* ASCII 121 'y'*/
      16, 17,
     "\x02\x0b\x08\x19\x7f\x0e\x0b\x08\x19\x06\x1d\x04\x1f"
      "\x02\x20\x01\x20"
   },
   { /* ASCII 122 'z'*/
      17, 14,
     "\x0e\x0b\x03\x19\x7f\x03\x0b\x0e\x0b\x7f\x03\x19"
      "\x0e\x19"
   },
   { /* ASCII 123 '{'*/
      14, 76,
     "\x09\x00\x07\x01\x06\x02\x05\x04\x05\x06\x06\x08\x07\x09"
      "\x08\x0b\x08\x0d\x06\x0f\x7f\x07\x01\x06\x03\x06\x05\x07\x07"
      "\x08\x08\x09\x0a\x09\x0c\x08\x0e\x04\x10\x08\x12\x09\x14\x09\x16"
      "\x08\x18\x07\x19\x06\x1b\x06\x1d\x07\x1f\x7f\x06\x11\x08\x13"
      "\x08\x15\x07\x17\x06\x18\x05\x1a\x05\x1c\x06\x1e\x07\x1f\x09\x20"
   },
   { /* ASCII 124 '|'*/
      8, 4,
     "\x04\x00\x04\x20"
   },
   { /* ASCII 125 '}'*/
      14, 76,
     "\x05\x00\x07\x01\x08\x02\x09\x04\x09\x06\x08\x08\x07\x09"
      "\x06\x0b\x06\x0d\x08\x0f\x7f\x07\x01\x08\x03\x08\x05\x07\x07"
      "\x06\x08\x05\x0a\x05\x0c\x06\x0e\x0a\x10\x06\x12\x05\x14\x05\x16"
      "\x06\x18\x07\x19\x08\x1b\x08\x1d\x07\x1f\x7f\x08\x11\x06\x13"
      "\x06\x15\x07\x17\x08\x18\x09\x1a\x09\x1c\x08\x1e\x07\x1f\x05\x20"
   },
   { /* ASCII 126 '~'*/
      24, 45,
     "\x03\x13\x03\x11\x04\x0e\x06\x0d\x08\x0d\x0a\x0e\x0e\x11"
      "\x10\x12\x12\x12\x14\x11\x15\x0f\x7f\x03\x11\x04\x0f\x06\x0e"
      "\x08\x0e\x0a\x0f\x0e\x12\x10\x13\x12\x13\x14\x12\x15\x0f\x15\x0d"
   },
};


void TekTerminal::tekChar ( int * pX, int * pY, char c, int fontsize  )
{ 
	if ( c > 31 && c < 127 )
	{
		int x_0, y_0;
		int x_1, y_1;
		int x_2, y_2;

		int idx = c - 32;
		byte first = 1;

		// reference point
		x_0 = *pX;
		y_0 = *pY;

		for ( int i=0 ; i<m_charset[idx].points ; )
		{
			BYTE x = m_charset[idx].code[i++];

			if ( x == 0x7F )
			{
				// next one is moveto
				first = 1;
			}
			else
			{
				// origin to baseline: - 25
				int y = m_charset[idx].code[i++] - 25;
				
				if ( first )
				{
					x_1 = x_0 + x * fontsize / 22;
					y_1 = y_0 + y * fontsize / 22;
					m_canvas->moveTo(x_1, y_1);
					first = 0;
				}
				else
				{
					// lineto
					x_2 = x_0 + x * fontsize / 22;
					y_2 = y_0 + y * fontsize / 22;
					m_canvas->lineTo(x_2, y_2);
					x_1 = x_2;
					y_1 = y_2;
				}
			}
	}
		// advance
		*pX += m_charset[idx].width * fontsize / 22; 
	}
}

void TekTerminal::soundBell()
{
	soundGenerator.attach(&sine);
	sine.enable(true);
	soundGenerator.setVolume(60);
	soundGenerator.play(true);
	sine.setFrequency(550);
	delay(75);
	soundGenerator.play(false);
	soundGenerator.detach(&sine);
}

void TekTerminal::eraseGraphics()
{
		int_clear();
		
		x_0 = 0;	// to upper left corner
		y_0 = cellheight / 11;	// character height
		
		xColumnStart = 0;	// reset ALPHA mode output to left column

		pen = 0;	// pen up
		seq = 0;
		getExtra = false;
		loX = 0;
		hiX = 0;
		loY = 0;
		hiY = 0;

		intensity = RGB888(0,255,0);	// reset to bright green
		grabIntensity = false;	// not in greenscale mode
}

// switch terminal to ANSI mode
void TekTerminal::setTextMode()
{
	// reset tektronix character height just in case 
	// to avoid starting next time with small font
	cellwidth  = 55;
	cellheight = 117;
	
	m_canvas->setPenColor(m_emuState.foregroundColor);

   setCursorPos(1, 1);
	mode = MODE_TEXT;
	intensity = RGB888(0,255,0);
	
	m_prevBlinkingTextEnabled = true;
	m_prevCursorEnabled = true;
}

// switch terminal to one of the GRAPHICS modes
void TekTerminal::setGraphicsMode(int newMode )
{
	mode = newMode;

	if ( mode == MODE_SPECIAL )
	{
		grabIntensity = true; // each point is preceded by an intensity byte
	}
	
	m_canvas->setPenColor(intensity);

	m_prevBlinkingTextEnabled = false;
	m_prevCursorEnabled       = false;
}

// override ANSI Terminal method
void TekTerminal::consumeInputQueue()
{
	char c = getNextCode(false);  // blocking call. false: do not process ctrl chars


	xSemaphoreTake(m_blinkTimerMutex, portMAX_DELAY);
	m_prevCursorEnabled = int_enableCursor(false);
	m_prevBlinkingTextEnabled = enableBlinkingText(false);

	if ( mode == MODE_TEXT )
	{
		// standard terminal emulation

		// ESC
		// Start escape sequence
		if (c == ASCII_ESC)
			consumeESC();

		else if (ISCTRLCHAR(c))
			execCtrlCode(c);

		else
		{
			if (m_emuState.characterSet[m_emuState.characterSetIndex] == 0 || (!m_emuState.ANSIMode && m_emuState.VT52GraphicsMode))
				c = DECGRAPH_TO_CP437[(uint8_t)c];
			setChar(c);
		}
	}
	else
	{
		if ( c == ASCII_ESC )
		{
			mode = MODE_ESC;
		}
		else if ( mode == MODE_ALPHA )
		{
			switch ( c )
			{
				case ASCII_BELL:
					soundBell();
					break;
				case ASCII_BS:
					x_0 -= cellwidth  / 11;
					break;
				case ASCII_HT:
					x_0 += cellwidth  / 11;
					break;
				case ASCII_LF:
					y_0 += cellheight / 11;
					if ( y_0 > 350 )
					{
						// jump to next column
						y_0 = cellheight / 11;
						x_0 += 350;
						if ( x_0 > 350 )
						{
							// beyond second column: back to left column
							x_0 = 0;
							// we do not clear the screen here - could be an option
						}
						xColumnStart = x_0;
					}
					break;
				case ASCII_VT:
					y_0 -= cellheight / 11;
					break;
				case ASCII_CR:
					x_0  = xColumnStart;
					break;
				case ASCII_FS:
					setGraphicsMode(MODE_POINTS);
					break;
				case ASCII_GS:
					setGraphicsMode(MODE_GRAPH); 
					seq = 0;
					pen = 0; // up
					break;
				case ASCII_RS:
					setGraphicsMode(MODE_INC);
					break;

				default:      
					if ( c > 31 && c < 127 )
					{
						// tekChar ( &x_0, &y_0, c, cellheight*2/3 / 11 );
						tekChar ( &x_0, &y_0, c, cellheight*5/6 / 11 );
					}
					break;
			}
		}
		else if ( mode == MODE_GRAPH || 
				    mode == MODE_POINTS || 
				    mode == MODE_SPECIAL )
		{
			switch ( c )
			{
				case ASCII_BELL:
					soundBell();
					break;

				case ASCII_FS:
					setGraphicsMode(MODE_POINTS);
					break;

				case ASCII_US:	
				case ASCII_CR:
					setGraphicsMode(MODE_ALPHA);
					break;

				case ASCII_GS:
					seq = 0;
					pen = 0;
					setGraphicsMode(MODE_GRAPH);
					break;

				case ASCII_RS:
					setGraphicsMode(MODE_INC);
					break;

				default:      
					if ( grabIntensity )
					{
						// first character in case of SPECIAL command
						// FABgl uses R8.G8.B8 format
						// green: 00000000.11111111.00000000 = 0x00FF00
						intensity = RGB888(0,c,0);
						grabIntensity = false;          // got it!
					}
					else if ( (c & 0x60) == 0x40 )   // loX: move or draw
					{
						loX = c;    // also triggers stroke
						
						/*
						 * enhanced mode:
						 * resolution = 4096 x 4096 points, (4096 x 3072 visible)
						 * 4096/7 x 3072/9 = 585 x 341
						 */

						// enhanced graphics mode scaled to 320 x 240 -> 462*346
						// the extra byte contains 4 bits of interest
						// 0000.00.11   2 least significant bits of x
						// 0000.11.00   2 least significant bits of y
						x_0 =  25 + ( ((hiX-32) << 7) | ((loX - 64) << 2) | (extra      & 0x03) ) / 7;
						y_0 = 346 - ( ((hiY-32) << 7) | ((loY - 96) << 2) | ((extra>>2) & 0x03) ) / 9;


						if ( mode == MODE_POINTS || mode == MODE_SPECIAL )
						{
							m_canvas->setPixel(x_0, y_0, intensity);
							//m_canvas->setPenColor(intensity);
							//m_canvas->moveTo(x_0, y_0);
							//m_canvas->lineTo(x_0, y_0);
						}
						else if ( pen == 1 )  //  mode == MODE_GRAPH
						{
							// lineto
							m_canvas->lineTo(x_0, y_0);
						}
						else
						{
							// moveto
							m_canvas->moveTo(x_0, y_0);
							// pen down
							pen = 1;
						}
						
						seq = 0;
						getExtra = false;
						if ( mode == MODE_SPECIAL )
						{
							// restart for next point
							grabIntensity = true;
						}
					}
					else if ( (c & 0x60) == 0x20 )   // hiY or hiX
					{
						if ( seq == 0 )   // hiY
						{
							hiY = c;
							seq = 1;
						}
						else              // hiX
						{
							hiX = c;
							seq = 0;
						}
					}
					else if ( (c & 0x60) == 0x60 )   // extra or loY
					{
						if ( getExtra == false )
						{
							getExtra = true;
						}
						else
						{
							extra = loY;     // we had an extra in loY, shift loY into extra
							getExtra = false;
						}
						loY = c;
						seq = 1; // next one must be hiX
					}
					else if ( c == 0x0D )   // CR
					{
						pen = 0;
						seq = 0;
						mode = MODE_ALPHA;
					}
					break;
			}
		}
		else if ( mode == MODE_GIN )
		{
			switch ( c )
			{
				case ASCII_CR:
					setGraphicsMode(MODE_ALPHA);
					break;
			}        
		}
		else if ( mode == MODE_INC )
		{
			switch ( c )
			{
				case ASCII_US:
				case ASCII_CR:
					setGraphicsMode(MODE_ALPHA);
					break;
			}        
		}
		else if ( mode == MODE_SPECIAL )
		{
			switch ( c )
			{
				case ASCII_RS:
					setGraphicsMode(MODE_INC);
					break;
				case ASCII_US:
				case ASCII_CR:
					setGraphicsMode(MODE_ALPHA);
					break;
			}
		}
		else if ( mode == MODE_ESC )
		{
			if ( c == ASCII_FF)
			{
				// ESC FF: Clear Screen
				eraseGraphics();
				extra = 0;
				setGraphicsMode(MODE_ALPHA);
			}
			else if ( c == ASCII_ETX )
			{
				// ESC ETX: go back to Text emulation
				setTextMode();
			}
			else if ( c == ASCII_SUB )
			{
				// ESC SUB
				setGraphicsMode(MODE_GIN);
			}
			else if ( c == ASCII_FS )
			{
				// ESC FS
				setGraphicsMode(MODE_SPECIAL);
			}
			else if ( c == '8' )	// 35 lines
			{
				// ESC '8' = 0x38 = 56d
				// default after startup
				cellwidth  = 55;
				cellheight = 117;
				setGraphicsMode(MODE_ALPHA);
			}
			else if ( c == '9' )	// 38 lines
			{
				// ESC '9' = 0x39 = 57d
				cellwidth  = 51;
				cellheight = 108;
				setGraphicsMode(MODE_ALPHA);
			}
			else if ( c == ':' )	// 58 lines
			{
				// ESC ':' = 0x3A = 58d
				cellwidth  = 34;
				cellheight = 71;
				setGraphicsMode(MODE_ALPHA);
			}
			else if ( c == ';' )	// 64 lines
			{
				// ESC ';' = 0x3B = 59d
				cellwidth  = 31;
				cellheight = 64;
				setGraphicsMode(MODE_ALPHA);
			}   
			else if ( c >= '`'  && c <= 'd')	// line type not used here
			{
				//           `,a,b,c,d 
				// line type 1,2,3,4,5
				// 1 = solid
				// 2 = dotted
				// 3 = dot-dash
				// 4 = short dash
				// 5 = long dash
				// translate into green levels [255...127]
				setGraphicsMode(MODE_ALPHA);
				intensity = RGB888(0,255 - 32 * (c - '\`'),0);
				m_canvas->setPenColor(intensity);
			}   
			else
			{
				setGraphicsMode(MODE_ALPHA);
				
				// handle escape sequence for text cursor movement
				consumeESC();
			}
		}
	}
	
	enableBlinkingText(m_prevBlinkingTextEnabled);
	int_enableCursor(m_prevCursorEnabled);
	xSemaphoreGive(m_blinkTimerMutex);

	if (m_resetRequested)
		reset();
}

void TekTerminal::execCtrlCode(char c)
{
	switch (c)
	{
	// BS
	// backspace, move cursor one position to the left (without wrap).
	case ASCII_BS:
		if (m_emuState.cursorX > 1)
			setCursorPos(m_emuState.cursorX - 1, m_emuState.cursorY);
		else if (m_emuState.reverseWraparoundMode)
		{
			int newX = m_columns;
			int newY = m_emuState.cursorY - 1;
			if (newY == 0)
				newY = m_rows;
			setCursorPos(newX, newY);
      }
      break;

	/* -MH- added sound for ^G BELL character */
	case ASCII_BELL:
		soundBell();
		break;

		
	// HT
	// Go to next tabstop or end of line if no tab stop
	case ASCII_HT:
		nextTabStop();
		break;

	// LF
	// Line feed
	case ASCII_LF:
		if (!m_emuState.cursorPastLastCol)
		{
			if (m_emuState.newLineMode)
				setCursorPos(1, m_emuState.cursorY);
			if (moveDown())
				scrollUp();
		}
	break;

	// VT (vertical tab)
	// FF (form feed)
	// Move cursor down
	case ASCII_VT:
	case ASCII_FF:
		if (moveDown())
			scrollUp();
		break;

	// CR
	// Carriage return. Move cursor at the beginning of the line.
	case ASCII_CR:
		setCursorPos(1, m_emuState.cursorY);
		break;

	// SO
	// Shift Out. Switch to Alternate Character Set (G1)
	case ASCII_SO:
		m_emuState.characterSetIndex = 1;
		break;

	// SI
	// Shift In. Switch to Standard Character Set (G0)
	case ASCII_SI:
		m_emuState.characterSetIndex = 0;
		break;

	case ASCII_DEL:
		// nothing to do
		break;

	default:
		break;
  }
}

// consume non-CSI and CSI (ESC is already consumed)
void TekTerminal::consumeESC()
{

  if (!m_emuState.ANSIMode)
  {
    consumeESCVT52();
    return;
  }

  char c = getNextCode(true);   // true: process ctrl chars

  if (c == '[') {
    // ESC [ : start of CSI sequence
    consumeCSI();
    return;
  }

  if (c == 'P') {
    // ESC P : start of DCS sequence
    consumeDCS();
    return;
  }

  #if FABGLIB_TERMINAL_DEBUG_REPORT_ESC
  logFmt("ESC%c\n", c);
  #endif

  switch (c)
  {
    // ESC c : RIS, reset terminal.
    case 'c':
      m_resetRequested = true;  // reset() actually called by consumeInputQueue()
      break;

    // ESC D : IND, line feed.
    case 'D':
      if (moveDown())
        scrollUp();
      break;

    // ESC E : NEL, new line.
    case 'E':
      setCursorPos(1, m_emuState.cursorY);
      if (moveDown())
        scrollUp();
      break;

    // ESC H : HTS, set tab stop at current column.
    case 'H':
      setTabStop(m_emuState.cursorX, true);
      break;

    // ESC M : RI, move up one line keeping column position.
    case 'M':
      if (moveUp())
        scrollDown();
      break;

    // ESC Z : DECID, returns TERMID
    case 'Z':
      //log("DECID\n");
      sendCSI();
      send(TERMID);
      break;

    // ESC 7 : DECSC, save current cursor state (position and attributes)
    case '7':
      saveCursorState();
      break;

    // ESC 8 : DECRC, restore current cursor state (position and attributes)
    case '8':
      restoreCursorState();
      break;

    // ESC #
    case '#':
      c = getNextCode(true);  // true: process ctrl chars
      switch (c) {
        // ESC # 3 : DECDHL, DEC double-height line, top half
        case '3':
          setLineDoubleWidth(m_emuState.cursorY, 2);
          break;
        // ESC # 4 : DECDHL, DEC double-height line, bottom half
        case '4':
          setLineDoubleWidth(m_emuState.cursorY, 3);
          break;
        // ESC # 5 : DECSWL, DEC single-width line
        case '5':
          setLineDoubleWidth(m_emuState.cursorY, 0);
          break;
        // ESC # 6 : DECDWL, DEC double-width line
        case '6':
          setLineDoubleWidth(m_emuState.cursorY, 1);
          break;
        // ESC # 8 :DECALN, DEC screen alignment test - fill screen with E's.
        case '8':
          erase(1, 1, m_columns, m_rows, 'E', false, false);
          break;
      }
      break;

    // Start sequence defining character set:
    // ESC character_set_index character_set
    // character_set_index: '(' = G0  ')' = G1   '*' = G2   '+' = G3
    // character_set: '0' = VT100 graphics mapping    '1' = VT100 graphics mapping    'B' = USASCII
    case '(':
    case ')':
    case '*':
    case '+':
      switch (getNextCode(true)) {
        case '0':
        case '2':
          m_emuState.characterSet[c - '('] = 0; // DEC Special Character and Line Drawing
          break;
        default:  // 'B' and others
          m_emuState.characterSet[c - '('] = 1; // United States (USASCII)
          break;
      }
      break;

    // ESC = : DECKPAM (Keypad Application Mode)
    case '=':
      m_emuState.keypadMode = KeypadMode::Application;
      #if FABGLIB_TERMINAL_DEBUG_REPORT_DESCSALL
      log("Keypad Application Mode\n");
      #endif
      break;

	// ESC > : DECKPNM (Keypad Numeric Mode)
	case '>':
		m_emuState.keypadMode = KeypadMode::Numeric;
		#if FABGLIB_TERMINAL_DEBUG_REPORT_DESCSALL
		log("Keypad Numeric Mode\n");
		#endif
		break;

	/* -MH- added for Tektronix */
	// ESC FF clear screen and go to ALPHA mode
	case ASCII_FF:
		eraseGraphics();
		extra = 0;
		setGraphicsMode(MODE_ALPHA);
      break;

    case ASCII_SPC:
      switch (getNextCode(true)) {

        // ESC SPC F : S7C1T, Select 7-Bit C1 Control Characters
        case 'F':
          m_emuState.ctrlBits = 7;
          break;

        // ESC SPC G : S8C1T, Select 8-Bit C1 Control Characters
        case 'G':
          if (m_emuState.conformanceLevel >= 2 && m_emuState.ANSIMode)
            m_emuState.ctrlBits = 8;
          break;
      }
      break;

    // consume unknow char
    default:
      #if FABGLIB_TERMINAL_DEBUG_REPORT_UNSUPPORT
      logFmt("Unknown ESC %c\n", c);
      #endif
      break;
  }
}


// consume DEC Private Mode (DECSET/DECRST) sequences
// ESC [ ? # h     <- set
// ESC [ ? # l     <- reset
// "c" can be "h" or "l"
void TekTerminal::consumeDECPrivateModes(int const * params, int paramsCount, char c)
{
  bool set = (c == 'h');
  switch (params[0]) {

    // ESC [ ? 1 h
    // ESC [ ? 1 l
    // DECCKM (default off): Cursor Keys Mode
    case 1:
      m_emuState.cursorKeysMode = set;
      break;

    // ESC [ ? 2 h
    // ESC [ ? 2 l
    // DECANM (default on): ANSI Mode (off = VT52 Mode)
    case 2:
      m_emuState.ANSIMode = set;
      //if (!set) Serial.write("VT52 Mode\n");
      break;

    // ESC [ ? 3 h
    // ESC [ ? 3 l
    // DECCOLM (default off): 132 Column Mode
    case 3:
      if (m_emuState.allow132ColumnMode) {
        set132ColumnMode(set);
        int_clear();
        setCursorPos(1, 1);
      }
      break;

    // ESC [ ? 4 h
    // ESC [ ? 4 l
    // DECSCLM (default off): Scrolling mode (jump or smooth)
    case 4:
      m_emuState.smoothScroll = set;
      break;


    // ESC [ ? 5 h
    // ESC [ ? 5 l
    // DECSCNM (default off): Reverse Video
    case 5:
      reverseVideo(set);
      break;

    // ESC [ ? 6 h
    // ESC [ ? 6 l
    // DECOM (default off): Origin mode
    case 6:
      m_emuState.originMode = set;
      if (set)
        setCursorPos(m_emuState.cursorX, m_emuState.scrollingRegionTop);
      break;

    // ESC [ ? 7 h
    // ESC [ ? 7 l
    // DECAWM (default on): Wraparound mode
    case 7:
      m_emuState.wraparound = set;
      break;

    // ESC [ ? 8 h
    // ESC [ ? 8 l
    // DECARM (default on): Autorepeat mode
    case 8:
      m_emuState.keyAutorepeat = set;
      break;

    // ESC [ ? 12 h
    // ESC [ ? 12 l
    // Start Blinking Cursor
    case 12:
      m_emuState.cursorBlinkingEnabled = set;
      break;

    // ESC [ ? 25 h
    // ESC [ ? 25 l
    // DECTECM (default on): Make cursor visible.
    case 25:
      m_prevCursorEnabled = set;  // consumeInputQueue() will set the value
      break;

	/* -MH- added for Tektronix */
	// ESC [ ? 38 h
	// ESC [ ? 38 l
	 // used to switch between Text and Tektronix emulation
    case 38:
		if ( set )
		{
			setGraphicsMode(MODE_ALPHA);
			eraseGraphics();
		}
		else
		{
			setTextMode();
		}
      break;

    // ESC [ ? 40 h
    // ESC [ ? 40 l
    // Allow 80 -> 132 column mode (default off)
    case 40:
      m_emuState.allow132ColumnMode = set;
      break;

    // ESC [ ? 45 h
    // ESC [ ? 45 l
    // Reverse-wraparound mode (default off)
    case 45:
      m_emuState.reverseWraparoundMode = set;
      break;

    // ESC [ ? 47 h
    // ESC [ ? 47 l
    // ESC [ ? 1047 h
    // ESC [ ? 1047 l
    // Use Alternate Screen Buffer
    case 47:
    case 1047:
      useAlternateScreenBuffer(set);
      break;

    // ESC [ ? 77 h
    // ESC [ ? 77 l
    // DECBKM (default off): Backarrow key Mode
    case 67:
      m_emuState.backarrowKeyMode = set;
      break;

    // ESC [ ? 1048 h  : Save cursor as in DECSC
    // ESC [ ? 1048 l  : Restore cursor as in DECSC
    case 1048:
      if (set)
        saveCursorState();
      else
        restoreCursorState();
      break;

    // ESC [ ? 1049 h
    // ESC [ ? 1049 l
    // Save cursor as in DECSC and use Alternate Screen Buffer
    case 1049:
      if (set) {
        saveCursorState();
        useAlternateScreenBuffer(true);
      } else {
        useAlternateScreenBuffer(false);
        restoreCursorState();
      }
      break;

    default:
      #if FABGLIB_TERMINAL_DEBUG_REPORT_UNSUPPORT
      logFmt("Unknown DECSET/DECRST: %d %c\n", params[0], c);
      #endif
      break;
  }
}


}